/*****************************************************************************/
/*!
 * \file theory.h
 * \brief Generic API for Theories plus methods commonly used by theories
 * 
 * Author: Clark Barrett
 * 
 * Created: Sat Nov 30 23:30:15 2002
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#ifndef _cvcl__include__theory_h_
#define _cvcl__include__theory_h_

#include <string>
#include <iostream>
#include "expr.h"
#include "expr_stream.h"
#include "theorem.h"
#include "type.h"
#include "common_proof_rules.h"
#include "cdlist.h"

namespace CVCL {

class VCL;

/************************************************************************/
/*!
 *\defgroup Theories Theories
 *\ingroup VC
 *\brief Theories
 *@{
 */
/***********************************************************************/

/*****************************************************************************/
/*!
 *\anchor Theory
 *\class Theory
 *\brief Base class for theories
 *
 * Author: Clark Barrett
 *
 * Created: Thu Jan 30 16:37:56 2003
 *
 * This is an abstract class which all theories should inherit from.  In
 * addition to providing an abstract theory interface, it provides access
 * functions to core functionality.  However, in order to avoid duplicating the
 * data structures which implement this functionality, all the functionality is
 * stored in a separate class (which actually derives from this one) called
 * TheoryCore.  These two classes work closely together to provide the core
 * functionality.
 */
/*****************************************************************************/
class Theory {
  friend class TheoryCore;
private:
  TheoryCore* d_theoryCore; //!< Provides the core functionality
  CommonProofRules* d_commonRules; //!< Commonly used proof rules
  std::string d_name; //!< Name of the theory (for debugging)

  //! Private default constructor.
  /*! Everyone besides TheoryCore has to use the public constructor
    which sets up all the provided functionality automatically.
  */
  Theory(void);

protected:
  bool d_theoryUsed; //! Whether theory has been used (for smtlib translator)

public:
  //! Exposed constructor.
  /*! Note that each instance of Theory must have a name (mostly for
    debugging purposes). */
  Theory(VCL* vcl, const std::string& name);
  //! Destructor
  virtual ~Theory(void);

  //! A command line flag whether to compute TCCs
  const bool* d_tccFlag;

  //! Get the name of the theory (for debugging purposes)
  const std::string& getName() const { return d_name; }

  //! Get whether theory has been used (for smtlib translator)
  virtual bool theoryUsed() { return d_theoryUsed; }

  //! Get a pointer to theoryCore
  TheoryCore* theoryCore() { return d_theoryCore; }

  /***************************************************************************/
  /*!
   *\defgroup Theory_API Abstract Theory Interface
   *\anchor theory_api
   *\ingroup Theories
   *\brief Abstract Theory Interface
   *
   * These are the theory-specific methods which provide the decision procedure
   * functionality for a new theory.  At the very least, a theory must
   * implement the checkSat method.  The other methods can be used to make the
   * implementation more convenient.  For more information on this API, see my
   * (Clark Barrett)'s PhD dissertation and \ref theory_api_howto.
   *@{
   */
  /***************************************************************************/

  //! Notify theory of a new shared term
  /*! When a term e associated with theory i occurs as a child of an expression
    associated with theory j, the framework calls i->addSharedTerm(e) and
    j->addSharedTerm(e)
  */
  virtual void addSharedTerm(const Expr& e) {}

  //! Assert a new fact to the decision procedure
  /*! Each fact that makes it into the core framework is assigned to exactly
    one theory: the theory associated with that fact.  assertFact is called to
    inform the theory that a new fact has been assigned to the theory.
  */
  virtual void assertFact(const Theorem& e) = 0;

  //! Receives all the type predicates for the types of the given theory
  /*!
   * Type predicates may be expensive to enqueue eagerly, and DPs may
   * choose to postpone them, or transform them to something more
   * efficient.  By default, the asserted type predicate is
   * immediately enqueued as a new fact.
   *
   * \param e is the expression for which the type predicate is computed
   * \param pred is the predicate theorem P(e)
   */
  virtual void assertTypePred(const Expr& e, const Theorem& pred)
    { enqueueFact(pred); }

  //! Check for satisfiability in the theory
  /*! \param fullEffort when it is false, checkSat can do as much or
   as little work as it likes, though simple inferences and checks for
   consistency should be done to increase efficiency.  If fullEffort is true,
   checkSat must check whether the set of facts given by assertFact together
   with the arrangement of shared terms (provided by addSharedTerm) induced by
   the global find database equivalence relation are satisfiable.  If
   satisfiable, checkSat does nothing.

   If satisfiability can be acheived by merging some of the shared terms, a new
   fact must be enqueued using enqueueFact (this fact need not be a literal).
   If there is no way to make things satisfiable, setInconsistent must be called.
  */
  virtual void checkSat(bool fullEffort) = 0;

  //! Add variables from 'e' to 'v' for constructing a concrete model
  /*! If e is already of primitive type, do NOT add it to v. */
  virtual void computeModelTerm(const Expr& e, std::vector<Expr>& v);
  //! Process disequalities from the arrangement for model generation
  virtual void refineCounterExample() {}
  //! Assign concrete values to basic-type variables in v
  virtual void computeModelBasic(const std::vector<Expr>& v) {}
  //! Compute the value of a compound variable from the more primitive ones
  /*! The more primitive variables for e are already assigned concrete
   * values, and are available through getModelValue().
   *
   * The new value for e must be assigned using assignValue() method.
   *
   * \param e is the compound type expression to assign a value;
   *
   * \param vars are the variables actually assigned.  Normally, 'e'
   * is the only element of vars.  However, e.g. in the case of
   * uninterpreted functions, assigning 'f' means assigning all
   * relevant applications of 'f' to constant values (f(0), f(5),
   * etc.).  Such applications might not be known before the model is
   * constructed (they may be of the form f(x), f(y+z), etc., where
   * x,y,z are still unassigned).
   *
   * Populating 'vars' is an opportunity for a DP to change the set of
   * top-level "variables" to assign, if needed.  In particular, it
   * may drop 'e' from the model entirely, if it is already a concrete
   * value by itself.
   */
  virtual void computeModel(const Expr& e, std::vector<Expr>& vars) {
    assignValue(find(e));
    vars.push_back(e);
  }


  //! Theory-specific rewrite rules.  
  /*! By default, rewrite just returns a reflexive theorem stating that the
    input expression is equivalent to itself.  However, rewrite is allowed to
    return any theorem which describes how the input expression is equivalent
    to some new expression.  rewrite should be used to perform simplifications,
    normalization, and any other preprocessing on theory-specific expressions
    that needs to be done.
  */
  virtual Theorem rewrite(const Expr& e) { return reflexivityRule(e); }
  //! Theory-specific rewrites for atomic formulas
  /*! The intended use is to convert complex atomic formulas into an
   * equivalent Boolean combination of simpler formulas.  Such
   * conversion may be harmful for algebraic rewrites, and is not
   * always desirable to have in rewrite() method.
   *
   * However, if rewrite() alone cannot solve the problem, and the SAT
   * solver needs to be envoked, these additional rewrites may ease
   * the job for the SAT solver.
   */
  virtual Theorem rewriteAtomic(const Expr& e) { return reflexivityRule(e); }
  //! Recursive simplification step
  /*!
   * INVARIANT: the result is a Theorem(e=e'), where e' is a fully
   * simplified version of e.  To simplify subexpressions recursively,
   * call simplify() function.
   *
   * This theory-specific method is called when the simplifier
   * descends top-down into the expression.  Normally, every kid is
   * simplified recursively, and the results are combined into the new
   * parent with the same operator (Op).  This functionality is
   * provided with the default implementation.
   *
   * However, in some expressions some kids may not matter in the
   * result, and can be skipped.  For instance, if the first kid in a
   * long AND simplifies to FALSE, then the entire expression
   * simplifies to FALSE, and the remaining kids do not need to be
   * simplified.
   *
   * This call is a chance for a DP to provide this type of
   * optimizations during the top-down phase of simplification.
   */
  virtual Theorem simplifyOp(const Expr& e);

  //! Set up the term e for call-backs when e or its children change.
  /*! setup is called once for each expression associated with the theory.  It
    is typically used to setup theory-specific data for an expression and to
    add call-back information for use with update.
    \sa update
  */
  virtual void setup(const Expr& e) {}

  //! Notify a theory of a new equality
  /*! update is a call-back used by the notify mechanism of the core theory.
    It works as follows.  When an equation t1 = t2 makes it into the core
    framework, the two find equivalence classes for t1 and t2 are merged.  The
    result is that t2 is the new equivalence class representative and t1 is no
    longer an equivalence class representative.  When this happens, the notify
    list of t1 is traversed.  Notify list entries consist of a theory and an
    expression d.  For each entry (i,d), i->update(e, d) is called, where e is
    the theorem corresponding to the equality t1=t2.

    To add the entry (i,d) to a term t1's notify list, a call must be made to
    t1.addNotify(i,d).  This is typically done in setup.

    \sa setup
  */
  virtual void update(const Theorem& e, const Expr& d) {}

  //! An optional solver.
  /*! The solve method can be used to implement a Shostak-style solver.  Since
    solvers do not in general combine, the following technique is used.  One
    theory is designated as the primary solver (in our case, it is the theory
    of arithmetic).  For each equation that enters the core framework, the
    primary solver is called to ensure that the equation is in solved form with
    respect to the primary theory.

    After the primary solver, the solver for the theory associated with the
    equation is called.  This solver can do whatever it likes, as long as the
    result is still in solved form with respect to the primary solver.  This is
    a slight generalization of what is described in my (Clark)'s PhD thesis.
  */
  virtual Theorem solve(const Theorem& e) { return e; }

  //! Notification of conflict
  /*!
   * Decision procedures implement this method when they want to be
   * notified about a conflict.
   * \param thm is the theorem of FALSE given to setInconsistent()
   */
  virtual void notifyInconsistent(const Theorem& thm) { }

  //! Compute and store the type of e
  /*!
   * \param e is the expression whose type is computed.  
   *
   * This function computes the type of the top-level operator of e,
   * and recurses into children using getType(), if necessary.
   */
  virtual void computeType(const Expr& e) {}
  //! Compute the base type of the top-level operator of an arbitrary type
  virtual Type computeBaseType(const Type& tp) { return tp; }

  /*! @brief  Theory specific computation of the subtyping predicate for 
   *  type t applied to the expression e.
   */ 
  /*! By default returns true. Each theory needs to compute subtype predicates
   *  for the types associated with it. So, for example, the theory of records
   *  will take a record type [# f1: T1, f2: T2 #] and an expression e
   *  and will return the subtyping predicate for e, namely:
   *  computeTypePred(T1, e.f1) AND computeTypePred(T2, e.f2)
   */ 
  virtual Expr computeTypePred(const Type& t, const Expr& e) 
    {return e.getEM()->trueExpr();}




  //! Compute and cache the TCC of e.
  virtual Expr computeTCC(const Expr& e);

  //! Theory-specific pretty-printing.
  /*! By default, print the top node in AST, and resume
    pretty-printing the children.  The same call e.print(os) can be
    used in DP-specific printers to use AST printing for the given
    node.  In fact, it is strongly recommended to add e.print(os) as
    the default for all the cases/kinds that are not handled by the
    particular pretty-printer.
  */
  virtual ExprStream& print(ExprStream& os, const Expr& e) {
    return e.print(os);
  }

  //! Theory-specific parsing implemented by the DP
  virtual Expr parseExprOp(const Expr& e) { return e; }
  /*@}*/ // End of Theory_API group

  /***************************************************************************/
  /*!
   *\name Theory Helper Methods
   * These methods provide basic functionality needed by all theories.
   *@{
   */
  /***************************************************************************/

  //! Access to ExprManager
  ExprManager* getEM();

  //! Register new kinds with the given theory
  void registerKinds(Theory* theory, std::vector<int>& kinds);
  //! Register a new theory
  void registerTheory(Theory* theory, std::vector<int>& kinds,
		      bool hasSolver=false);

  //! Return the number of registered theories
  int getNumTheories();

  //! Return the theory associated with a kind
  Theory* theoryOf(int kind);
  //! Return the theory associated with an Expr
  Theory* theoryOf(const Expr& e);

  //! Return the theorem that e is equal to its find
  Theorem find(const Expr& e);

  //! Return the find of e, or e if it has no find
  inline const Expr& findExpr(const Expr& e)
  {
    return e.hasFind() ? find(e).getRHS() : e;
  }

  //! Calls the correct theory to compute a type predicate
  Expr getTypePred(const Type& t, const Expr& e);

  /*! @brief Calls the correct theory to get all of the terms that
    need to be assigned values in the concrete model */
  void getModelTerm(const Expr& e, std::vector<Expr>& v);
  //! Fetch the concrete assignment to the variable during model generation
  Theorem getModelValue(const Expr& e);

  //! Compute (if necessary) the type of e, and return the result
  const Type& getType(const Expr& e);
  //! Compute the TCC of e, or the subtyping predicate, if e is a type
  const Expr& getTCC(const Expr& e);
  //! Compute (or look up in cache) the base type of e and return the result
  Type getBaseType(const Expr& e);
  //! Compute the base type from an arbitrary type
  Type getBaseType(const Type& tp);

  //! Parse the generic expression. 
  /*! This method should be used in parseExprOp() for recursive calls
   *  to subexpressions, and is the method called by the command
   *  processor.
   */
  virtual Expr parseExpr(const Expr& e);
  //! Update the children of the term e
  /*! When a decision procedure receives a call to update() because a
    child of a term 'e' has changed, this method can be called to
    compute the new value of 'e'.
    \sa update
  */
  Theorem updateHelper(const Expr& e);
  //! Setup additional terms within a theory-specific setup().
  void setupTerm(const Expr& e);
  //! Setup additional terms within a theory-specific setup().
  /*!
   * \param e is the expression to setup (subterms will be setup recursively)
   * \param i is a pointer to a theory, normally the calling theory ('this')
   */
  virtual void setupTerm(const Expr& e, Theory* i);
  //! Setup a term for congruence closure (must have sig and rep attributes)
  void setupCC(const Expr& e);
  //! Update a term w.r.t. congruence closure (must be setup with setupCC())
  void updateCC(const Theorem& e, const Expr& d);
  //! Rewrite a term w.r.t. congruence closure (must be setup with setupCC())
  Theorem rewriteCC(const Expr& e);
  //! Create an operator (class Op) out of a kind
  Op getOpKind(int kind);

  /*@}*/ // End of Theory Helper Methods

  /***************************************************************************/
  /*!
   *\name Core Framework Functionality
   * The core methods for cooperating decision procedures.  These are the most
   * basic methods for decision procedures and the search engine to notify the
   * core framework about new facts, and also for polling the framework about
   * the state of the system (e.g. whether the context is inconsistent).
   *@{
   */
  /***************************************************************************/

  /*! @brief Add a new assertion to the core from the user or a SAT
    solver.  Do NOT use it in a decision procedure; use
    enqueueFact(). */
  /*! \sa enqueueFact */
  virtual void addFact(const Theorem& e);

  //! Simplify a term e w.r.t. the current context
  /*! \sa simplifyThm */
  Expr simplifyExpr(const Expr& e, bool forceRebuild = false)
    { return simplifyThm(e, forceRebuild).getRHS(); }

  //! Check if the current context is inconsistent
  virtual bool inconsistent();
  //! Check if the current decision branch is marked as incomplete
  virtual bool incomplete();
  //! Check if the branch is incomplete, and return the reasons (stings)
  virtual bool incomplete(std::vector<std::string>& reasons);

  //! Handle new equalities (usually asserted through addFact)
  /*!
   * INVARIANT: the Theorem 'e' is an equality e1==e2, where e2 is
   * i-leaf simplified in the current context, or a conjunction of
   * such equalities.
   *
   * IMPORTANT: decision procedures should NOT use this function in
   * update(); use enqueueEquality() instead.
   */
  virtual void assertEqualities(const Theorem& e);

  //! Simplify a term e and return a Theorem(e==e')
  /*! \sa simplifyExpr() */
  virtual Theorem simplifyThm(const Expr& e, bool forceRebuild = false);
  //! Recursive call of the simplifier.  See also simplifyOp()
  virtual Theorem simplify(const Expr& e);

  /*! @brief Returns phi |= e = rewriteCore(e).  "Core" added to avoid
    conflict with theory interface function name */
  virtual Theorem rewriteCore(const Expr& e);
  //! Top-level rewrites for literals.  See also rewriteAtomic()
  /*! Intended to be used by the SearchEngine. */
  virtual Theorem rewriteLiteral(const Expr& e);

  //! Submit a derived fact to the core from a decision procedure
  /*!
   * \param e is the Theorem for the new fact 
   */
  virtual void enqueueFact(const Theorem& e);
  //! Enqueue a new equality e: e1==e2, where e2 is i-leaf simplified
  void enqueueEquality(const Theorem& e);

  //! Make the context inconsistent; e must be Theorem(FALSE)
  virtual void setInconsistent(const Theorem& e);
  //! Suggest a splitter to the SearchEngine
  virtual void addSplitter(const Expr& e, int priority = 0);
  //! Mark the current decision branch as possibly incomplete
  /*!
   * This should be set when a decision procedure uses an incomplete
   * algorithm, and cannot guarantee satisfiability after the final
   * checkSat() call with full effort.  An example would be
   * instantiation of universal quantifiers.
   *
   * A decision procedure can provide a reason for incompleteness,
   * which will be reported back to the user.
   */
  virtual void setIncomplete(const std::string& reason);

  //! Get the proof of inconsistency for the current context
  /*! \return Theorem(FALSE) */
  virtual Theorem inconsistentThm();

  /*! @brief To be called by SearchEngine when it believes the context
    is satisfiable. */
  /*! \return true if NO new facts were generated */
  virtual bool checkSATCore();

  //! Set the resource limit (0==unlimited).
  virtual void setResourceLimit(unsigned limit);  

  //! Convenient method for chaining rewrites
  /*! \param thm is a Theorem(e1==e2)

    It runs the corresponding theory method on e2, gets a new
    Theorem(e2==e3), and applies transitivity to obtain Theorem(e1==e3).
    Theories should NOT redefine this method.

    \return Theorem(e1==e3)
  */
  Theorem rewriteThm(const Theorem& thm);

  /*@}*/ // End of Core Framework Functionality

  /***************************************************************************/
  /*!
   *\name Core Testers
   *@{
   */
  /***************************************************************************/

  //! Test if e is a term (as opposed to a predicate/formula)
  bool isTerm(const Expr& e) { return !getType(e).isBool(); }

  //! Test if e is a literal
  /*! A literal is an atomic formula, or its negation.  
    \sa isAtomicFormula */
  bool isLiteral(const Expr& e)
    { return (isAtomicFormula(e) || (e.isNot() && isAtomicFormula(e[0]))); }

  //! Test if e is an atomic term or a formula.
  /*! An atomic term is a term that does not contain a formula inside it.
      \sa isAtomicFormula */
  bool isAtomic(const Expr& e);

  //! Test if e is an atomic formula
  /*! An atomic formula is a predicate or a variable which does not
    contain any formula inside it.  For instance, a predicate "x = IF
    f THEN y ELSE z ENDIF is not an atomic forumula, since it contains
    the conditional "f", which is a formula. */
  bool isAtomicFormula(const Expr& e);

  //! Test if e is an i-leaf term for the current theory
  /*! A term 'e' is an i-leaf for a theory 'i', if it is a variable,
    or 'e' belongs to a different theory.  This definition makes sense
    for a larger term which by itself belongs to the current theory
    'i', but (some of) its children are variables or belong to
    different theories. */
  bool isLeaf(const Expr& e) { return e.isVar() || theoryOf(e) != this; }

  //! Test if e1 is an i-leaf in e2
  /*! \sa isLeaf */
  bool isLeafIn(const Expr& e1, const Expr& e2);

  //! Test if all i-leaves of e are simplified
  /*! \sa isLeaf */
  bool leavesAreSimp(const Expr& e);

  /*@}*/ // End of Core Testers

  // Core type constructors
  //! Return BOOLEAN type
  Type boolType();

  // Core expr constructors
  //! Return FALSE Expr
  const Expr& falseExpr();
  //! Return TRUE Expr
  const Expr& trueExpr();
  //! Create a new variable given its name and type
  /*! Add the variable to the database for resolving IDs in parseExpr
   */
  Expr newVar(const std::string& name, const Type& type);
  //! Create a new named expression given its name, type, and definition
  /*! Add the definition to the database for resolving IDs in parseExpr
   */
  Expr newVar(const std::string& name, const Type& type, const Expr& def);
  //! Create a new uninterpreted function
  /*! Add the definition to the database for resolving IDs in parseExpr
   */
  Op newFunction(const std::string& name, const Type& type);
  //! Create a new defined function
  /*! Add the definition to the database for resolving IDs in parseExpr
   */
  Op newFunction(const std::string& name, const Type& type, const Expr& def);
  //! Create a new bound variable given its name, uid and type
  /*! Does NOT add the variable to the parsing database.  Use
   * addBoundVar() in parseExpr() and parseExprOp().
   */
  Expr newBoundVar(const std::string& name, const std::string& uid,
		   const Type& type);
  //! Create and add a new bound variable to the stack, for parseExprOp().
  /*!
   * The stack is popped automatically upon return from the
   * parseExprOp() which used this method.
   *
   * Bound variable names may repeat, in which case the latest
   * declaration takes precedence.
   */
  Expr addBoundVar(const std::string& name, const Type& type);
  //! Create and add a new bound named def to the stack, for parseExprOp().
  /*!
   * The stack is popped automatically upon return from the
   * parseExprOp() which used this method.
   *
   * Bound variable names may repeat, in which case the latest
   * declaration takes precedence.
   *
   * The type may be Null, but 'def' must always be a valid Expr
   */
  Expr addBoundVar(const std::string& name, const Type& type, const Expr& def);
  /*! @brief Lookup variable and return it and its type.  Return NULL Expr if
    it doesn't exist yet. */
  Expr lookupVar(const std::string& name, Type* type);
  //! Create a new uninterpreted type with the given name
  /*! Add the name to the global variable database d_globals
   */
  Type newTypeExpr(const std::string& name);
  //! Create a new type abbreviation with the given name 
  Type newTypeExpr(const std::string& name, const Type& def);
  //! Resolve an identifier, for use in parseExprOp()
  /*!
   * First, search the bound variable stack, and if the name is not
   * found, search the global constant and type declarations.  
   *
   * \return an expression to use in place of the identifier, or Null
   * if cannot resolve the name.
   */
  Expr resolveID(const std::string& name);
  //! Return a reference to the list of all terms
  virtual const CDList<Expr>& getTerms();

  /***************************************************************************/
  /*!
   *\name Core Proof Rules
   *\anchor theory_api_core_proof_rules
   *@{
   */
  /***************************************************************************/

  //!  ==> a == a
  Theorem reflexivityRule(const Expr& a)
    { return d_commonRules->reflexivityRule(a); }
  //!  a1 == a2 ==> a2 == a1
  Theorem symmetryRule(const Theorem& a1_eq_a2)
    { return d_commonRules->symmetryRule(a1_eq_a2); }
  //! ==> (a1 == a2) == (a2 == a1)
  Theorem rewriteUsingSymmetry(const Expr& a1_eq_a2)
    { return d_commonRules->rewriteUsingSymmetry(a1_eq_a2); }
  //! (a1 == a2) & (a2 == a3) ==> (a1 == a3)
  Theorem transitivityRule(const Theorem& a1_eq_a2,
			   const Theorem& a2_eq_a3)
    { return d_commonRules->transitivityRule(a1_eq_a2, a2_eq_a3); }
  //! (c_1 == d_1) & ... & (c_n == d_n) ==> op(c_1,...,c_n) == op(d_1,...,d_n)
  Theorem substitutivityRule(const Op& op,
			     const std::vector<Theorem>& thms)
    { return d_commonRules->substitutivityRule(op, thms); }

  //! Optimized: only positions which changed are included
  Theorem substitutivityRule(const Expr& e,
			     const std::vector<unsigned>& changed,
			     const std::vector<Theorem>& thms)
    { return d_commonRules->substitutivityRule(e, changed, thms); }

  //! |- e,  |- !e ==> |- FALSE
  /*!
   * INVARIANT: must generate exactly 2 assumptions for its theorem
   * (SearchEngineFast relies on that).
   */
  Theorem contradictionRule(const Theorem& e, const Theorem& not_e) {
    return d_commonRules->contradictionRule(e, not_e);
  }

  //! e ==> e IFF TRUE
  Theorem iffTrue(const Theorem& e) {
    return d_commonRules->iffTrue(e);
  }

  //! e ==> !e IFF FALSE
  Theorem iffNotFalse(const Theorem& e) {
    return d_commonRules->iffNotFalse(e);
  }

  //! e IFF TRUE ==> e
  Theorem iffTrueElim(const Theorem& e) {
    return d_commonRules->iffTrueElim(e);
  }

  //! !e ==> e IFF FALSE
  Theorem notToIff(const Theorem& not_e) {
    return d_commonRules->notToIff(not_e);
  }


  //! e IFF FALSE ==> !e
  Theorem iffFalseElim(const Theorem& e) {
    return d_commonRules->iffFalseElim(e);
  }

  //! e1 <=> e2  ==>  ~e1 <=> ~e2
  /*!  \f[\frac{\Gamma\vdash e_1\Leftrightarrow e_2}
   *           {\Gamma\vdash\sim e_1\Leftrightarrow\sim e_2}\f]
   * Where ~e is the <em>inverse</em> of e (that is, ~(!e') = e').
   */
  Theorem iffContrapositive(const Theorem& thm) {
    return d_commonRules->iffContrapositive(thm);
  }

  //! !!e ==> e
  Theorem notNotElim(const Theorem& not_not_e) {
    return d_commonRules->notNotElim(not_not_e);
  }

  //! e1 AND (e1 IFF e2) ==> e2
  Theorem iffMP(const Theorem& e1, const Theorem& e1_iff_e2) {
    return d_commonRules->iffMP(e1, e1_iff_e2);
  }

  //! AND(e_1,...e_n) ==> e_i
  Theorem andElim(const Theorem& e, int i) {
    return d_commonRules->andElim(e, i);
  }

  //! e1, e2 ==> AND(e1, e2)
  Theorem andIntro(const Theorem& e1, const Theorem& e2) {
    return d_commonRules->andIntro(e1, e2);
  }

  //! e1, ..., en ==> AND(e1, ..., en)
  Theorem andIntro(const std::vector<Theorem>& es) {
    return d_commonRules->andIntro(es);
  }

  //! ==> (e1 <=> e2) <=> [simplified expr]
    /*! Rewrite formulas like FALSE/TRUE <=> e,  e <=> NOT e, etc. */
  Theorem rewriteIff(const Expr& e) {
    return d_commonRules->rewriteIff(e);
  }

  //! ==> AND(e1,e2) IFF [simplified expr]
  Theorem rewriteAnd(const Expr& e) {
    return d_commonRules->rewriteAnd(e);
  }
  //! ==> OR(e1,...,en) IFF [simplified expr]
  Theorem rewriteOr(const Expr& e) {
    return d_commonRules->rewriteOr(e);
  }
  
  // Derived rules

  Theorem trueTheorem()
  {
    return iffTrueElim(reflexivityRule(trueExpr()));
  }

  //! AND(e1,e2) ==> [simplified expr]
  Theorem rewriteAnd(const Theorem& e) {
    return iffMP(e, d_commonRules->rewriteAnd(e.getExpr()));
  }
  //! OR(e1,...,en) ==> [simplified expr]
  Theorem rewriteOr(const Theorem& e) {
    return iffMP(e, d_commonRules->rewriteOr(e.getExpr()));
  }

  //! ==> NOT TRUE IFF FALSE
  Theorem rewriteNotTrue(const Expr& e) {
    return d_commonRules->rewriteNotTrue(e);
  }

  //! ==> NOT FALSE IFF TRUE
  Theorem rewriteNotFalse(const Expr& e) {
    return d_commonRules->rewriteNotFalse(e);
  }

  //! ==> NOT NOT e IFF e, takes e
  Theorem rewriteNotNot(const Expr& e) {
    return d_commonRules->rewriteNotNot(e);
  }

  //! ==> NOT FORALL (vars): e  IFF EXISTS (vars) NOT e
  Theorem rewriteNotForall(const Expr& forallExpr) {
    return d_commonRules->rewriteNotForall(forallExpr);
  }

  //! ==> NOT EXISTS (vars): e  IFF FORALL (vars) NOT e
  Theorem rewriteNotExists(const Expr& existsExpr) {
    return d_commonRules->rewriteNotExists(existsExpr);
  }

  /*! @brief 
  ==> u:a |- a, where u is optional proof label.  If not given, a new
    unique proof label is generated. */
  Theorem assumpRule(const Expr& a)
    { return d_commonRules->assumpRule(a); }

  Theorem rewriteIteThen(const Expr& e, const Theorem& thenThm){
  return d_commonRules->rewriteIteThen(e, thenThm); }

  Theorem rewriteIteElse(const Expr& e, const Theorem& elseThm){
  return d_commonRules->rewriteIteElse(e, elseThm); }

  // ==> NOT(e) == ITE(e, FALSE, TRUE)
  Theorem NotToIte(const Expr& e){
    return d_commonRules->NotToIte(e);
  }

  // ==> Or(e) == ITE(e[1], TRUE, e[0])
  Theorem OrToIte(const Expr& e){
    return d_commonRules->OrToIte(e);
  }

  // ==> And(e) == ITE(e[1], e[0], FALSE)
  Theorem AndToIte(const Expr& e){
    return d_commonRules->AndToIte(e);
  }

  // ==> IFF(a,b) == ITE(a, b, !b)
  Theorem IffToIte(const Expr& e){
    return d_commonRules->IffToIte(e);
  }

  // ==> IMPLIES(a,b) == ITE(a, b, TRUE)
  Theorem ImpToIte(const Expr& e){
    return d_commonRules->ImpToIte(e);
  }

  //! e: T ==> |- typePred_T(e)  [deriving the type predicate of e]
  Theorem typePred(const Expr& e) {
    return d_commonRules->typePred(e);
  }

  //! |- EXISTS x. e = x
  Theorem varIntroRule(const Expr& phi) {
    return d_commonRules->varIntroRule(phi);
  }

  //! |- P(ITE(cond,a,b)) <=> ITE(cond,Pred(a),Pred(b))
  Theorem ifLiftRule(const Expr& e, unsigned int itePos) {
    return d_commonRules->ifLiftRule(e, itePos);
  }

  //! AND(x1,...,xn) <=> v  |-  CNF[AND(x1,...,xn) <=> v]
  Theorem andCNFRule(const Theorem& thm) {
    return d_commonRules->andCNFRule(thm);
  }
  
  //! OR(x1,...,xn) <=> v  |-  CNF[OR(x1,...,xn) <=> v]
  Theorem orCNFRule(const Theorem& thm) {
    return d_commonRules->orCNFRule(thm);
  }

  //! (x1 <=> x2) <=> v  |-  CNF[(x1 <=> x2) <=> v]
  Theorem impCNFRule(const Theorem& thm) {
    return d_commonRules->impCNFRule(thm);
  }

  //! (x1 => x2) <=> v  |-  CNF[(x1 => x2) <=> v]
  Theorem iffCNFRule(const Theorem& thm) {
    return d_commonRules->iffCNFRule(thm);
  }

  //! ITE(c, x1, x2) <=> v  |-  CNF[ITE(c, x1, x2) <=> v]
  Theorem iteCNFRule(const Theorem& thm) {
    return d_commonRules->iteCNFRule(thm);
  }

  //! ITE(c, f1, f2) |- (NOT c OR f1) AND (c OR f2)
  Theorem iteToClauses(const Theorem& ite) {
    return d_commonRules->iteToClauses(ite);
  }

  //! e1 <=> e2 |- (NOT e1 OR e2) AND (e1 OR NOT e2)
  Theorem iffToClauses(const Theorem& ite) {
    return d_commonRules->iffToClauses(ite);
  }

  //! e: T ==> |- [all the subtype-related facts for e and its subterms]
  Theorem subtypePredicate(const Expr& e);

  /*! @brief Convert 2-valued formula to 3-valued by discharging its
   *  TCC (\f$D_\phi\f$):
   *  \f[\frac{\Gamma_1\vdash_2 \phi\quad \Gamma_2\vdash_2 D_{\phi}}
   *          {\Gamma_1,\,\Gamma_2\vdash_3\phi}\f]
   */
  Theorem3 queryTCC(const Theorem& phi, const Theorem& D_phi) {
    return d_commonRules->queryTCC(phi, D_phi);
  }

  //  G0,a1,...,an |-_3 phi  G1 |- D_a1 ... Gn |- D_an
  // -------------------------------------------------
  //    G0,G1,...,Gn |-_3 (a1 & ... & an) -> phi
  /*!
   * @brief 3-valued implication introduction rule:
   * \f[\frac{\Gamma_0,\,\alpha_1\,\ldots,\,\alpha_n\vdash_3\phi\quad
   *          (\Gamma_i\vdash D_{\alpha_i})_{i\in[0..n]}}
   *         {\Gamma_1,\,\Gamma_2\vdash_3
   *              (\bigwedge_{i=1}^n\alpha_i)\to\phi}\f]
   *
   * \param phi is the formula \f$\phi\f$
   * \param assump is the vector of assumptions \f$\alpha_1\ldots\alpha_n\f$
   * \param tccs is the vector of TCCs for assumptions
   *   \f$D_{\alpha_1}\ldots D_{\alpha_n}\f$
   */
  Theorem3 implIntro3(const Theorem3& phi,
		      const std::vector<Expr>& assump,
		      const std::vector<Theorem>& tccs) {
    return d_commonRules->implIntro3(phi, assump, tccs);
  }

  //! exists x. phi(x) ==> phi(a), where 'a' is a Skolem constant
  virtual Theorem skolemize(const Theorem& thm);
  //! Returns a theorem |- e = v for a new Skolem variable v
  virtual Theorem varIntroSkolem(const Expr& e);
  //! Assigns t a concrete value val.  Used in model generation.
  virtual void assignValue(const Expr& t, const Expr& val);
  //! Record a derived assignment to a variable (LHS).
  virtual void assignValue(const Theorem& thm);
  /*@}*/ // End of Core Proof Rules

};

/*@}*/ // End of group Theories

}

#endif
