/*****************************************************************************/
/*!
 * \file common_theorem_producer.h
 * 
 * Author: Sergey Berezin
 * 
 * Created: Feb 05 03:40:36 GMT 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/
// CLASS: CommonTheoremProducer
//
// AUTHOR: Sergey Berezin, 12/09/2002
//
// Description: Implementation of the proof rules for ground
// equational logic (reflexivity, symmetry, transitivity, and
// substitutivity).
//
// The method newTheorem() is inherited from TheoremProducer and,
// therefore, is not listed in this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _CVC_lite__common_theorem_producer_h_
#define _CVC_lite__common_theorem_producer_h_

#include "theorem.h"
#include "common_proof_rules.h"
#include "theorem_producer.h"
#include "expr_map.h"

namespace CVCL {

  class TheoryCore;
  
  class CommonTheoremProducer: 
    public CommonProofRules,
    public TheoremProducer {
  private:
    //! pointer to theory core
    TheoryCore* d_core;
  public:
    // Constructor
    CommonTheoremProducer(TheoremManager* tm, TheoryCore* core)
      : TheoremProducer(tm), d_core(core) { }
    // Destructor
    virtual ~CommonTheoremProducer() { }
    // ==> u:a |- a
    Theorem assumpRule(const Expr& a);
    //  ==> a == a
    Theorem reflexivityRule(const Expr& a);
    //  a1 == a2 ==> a2 == a1 (same for IFF)
    Theorem symmetryRule(const Theorem& a1_eq_a2);
    // ==> (a1 == a2) IFF (a2 == a1)
    Theorem rewriteUsingSymmetry(const Expr& a1_eq_a2);
    // (a1 == a2) & (a2 == a3) ==> (a1 == a3)   [same for IFF]
    Theorem transitivityRule(const Theorem& a1_eq_a2,
			     const Theorem& a2_eq_a3);
    // (c_1 == d_1) & ... & (c_n == d_n)
    //   ==> op(c_1,...,c_n) == op(d_1,...,d_n)
    Theorem substitutivityRule(const Op& op,
			       const std::vector<Theorem>& thms);
    Theorem substitutivityRule(const Expr& e,
			       const std::vector<unsigned>& changed,
			       const std::vector<Theorem>& thms);

    ////////////////////////////////////////////////////////////////////////
    // TCC rules (3-valued logic)
    ////////////////////////////////////////////////////////////////////////

    //  G1 |- phi   G2 |- D_phi
    // -------------------------
    //       G1,G2 |-_3 phi
    Theorem3 queryTCC(const Theorem& phi, const Theorem& D_phi);

    //  G0,a1,...,an |-_3 phi  G1 |- D_a1 ... Gn |- D_an
    // -------------------------------------------------
    //    G0,G1,...,Gn |-_3 (a1 & ... & an) -> phi
    Theorem3 implIntro3(const Theorem3& phi,
			const std::vector<Expr>& assump,
			const std::vector<Theorem>& tccs);

    ////////////////////////////////////////////////////////////////////////
    // Core derivation rules
    ////////////////////////////////////////////////////////////////////////

    // |- e,  |- !e ==> |- FALSE
    Theorem contradictionRule(const Theorem& e, const Theorem& not_e);
    // e ==> e IFF TRUE
    Theorem iffTrue(const Theorem& e);
    // e ==> !e IFF FALSE
    Theorem iffNotFalse(const Theorem& e);
    // e IFF TRUE ==> e
    Theorem iffTrueElim(const Theorem& e);
    // e IFF FALSE ==> !e
    Theorem iffFalseElim(const Theorem& e);
    // e1 <=> e2  ==>  ~e1 <=> ~e2
    Theorem iffContrapositive(const Theorem& thm);
    // !!e ==> e
    Theorem notNotElim(const Theorem& e);
    // e1 AND (e1 IFF e2) ==> e2
    Theorem iffMP(const Theorem& e1, const Theorem& e1_iff_e2);
    // e1 AND (e1 IMPLIES e2) ==> e2
    Theorem implMP(const Theorem& e1, const Theorem& e1_impl_e2);
    // AND(e_0,...e_{n-1}) ==> e_i
    Theorem andElim(const Theorem& e, int i);
    // e1, e2 ==> AND(e1, e2)
    Theorem andIntro(const Theorem& e1, const Theorem& e2);
    Theorem andIntro(const std::vector<Theorem>& es);
    // e: T ==> |- typePred_T(e)  [asserting the type predicate of e]
    Theorem typePred(const Expr& e);
    
    //  G,a1,...,an |- phi
    // -------------------------------------------------
    //    G |- (a1 & ... & an) -> phi
    /*!
     * @brief Implication introduction rule:
     * \f[\frac{\Gamma,\,\alpha_1\,\ldots,\,\alpha_n\vdash\phi}
     *         {\Gamma\vdash(\bigwedge_{i=1}^n\alpha_i)\to\phi}\f]
     *
     * \param phi is the formula \f$\phi\f$
     * \param assump is the vector of assumptions \f$\alpha_1\ldots\alpha_n\f$
     */
    Theorem implIntro(const Theorem& phi, const std::vector<Expr>& assump);

    //! e1 => e2  ==>  ~e2 => ~e1
    /*!  \f[\frac{\Gamma\vdash e_1\Rightarrow e_2}
     *           {\Gamma\vdash\sim e_2\Rightarrow\sim e_1}\f]
     * Where ~e is the <em>inverse</em> of e (that is, ~(!e') = e').
     */
    Theorem implContrapositive(const Theorem& thm);

    ////////////////////////////////////////////////////////////////////////
    // Derived rules                                                      //
    ////////////////////////////////////////////////////////////////////////

    // NOT e ==> e IFF FALSE
    Theorem notToIff(const Theorem& not_e);

    ////////////////////////////////////////////////////////////////////////
    // Core rewrite rules
    ////////////////////////////////////////////////////////////////////////

    // Rewrite rules used only in TheoryCore to replace LETDECL and
    // CONSTDEF with their definitions
    Theorem rewriteLetDecl(const Expr& e);
    Theorem rewriteConstDef(const Expr& e);
    // ==> (a == a) IFF TRUE
    Theorem rewriteReflexivity(const Expr& a_eq_a);
    // ==> NOT TRUE IFF FALSE
    Theorem rewriteNotTrue(const Expr& e);
    // ==> NOT FALSE IFF TRUE
    Theorem rewriteNotFalse(const Expr& e);
    // ==> NOT NOT e IFF e, takes !!e
    Theorem rewriteNotNot(const Expr& e);
    // ==> NOT (AND e1 ... en) IFF (OR !e1 ... !en), takes (AND ...)
    Theorem rewriteNotAnd(const Expr& e);
    // ==> NOT (OR e1 ... en) IFF (AND !e1 ... !en), takes (OR ...)
    Theorem rewriteNotOr(const Expr& e);
    // ==> NOT IFF(e1,e2) IFF IFF(e1,NOT e2)
    Theorem rewriteNotIff(const Expr& e);
    // ==> NOT ITE(a,b,c) IFF ITE(a,NOT b,NOT c)
    Theorem rewriteNotIte(const Expr& e);
    //! ==> NOT FORALL (vars): e  IFF EXISTS (vars) NOT e
    Theorem rewriteNotForall(const Expr& forallExpr);
    //! ==> NOT EXISTS (vars): e  IFF FORALL (vars) NOT e
    Theorem rewriteNotExists(const Expr& existsExpr);
    // ==> ITE(TRUE, e1, e2) == e1
    Theorem rewriteIteTrue(const Expr& e);
    // ==> ITE(FALSE, e1, e2) == e2
    Theorem rewriteIteFalse(const Expr& e);
    // ==> ITE(c, e, e) == e
    Theorem rewriteIteSame(const Expr& e);
    // ==> ITE(c, e1, e2) <=> (c => e1) AND (!c => e2)
    Theorem rewriteIteBool(const Expr& c,
			   const Expr& e1, const Expr& e2);
    // a |- b == d ==> ITE(a, b, c) == ITE(a, d, c)
    Theorem rewriteIteThen(const Expr& e, const Theorem& thenThm);
    // !a |- c == d ==> ITE(a, b, c) == ITE(a, b, d)
    Theorem rewriteIteElse(const Expr& e, const Theorem& elseThm);

    // AND and OR rewrites check for TRUE and FALSE arguments and
    // remove them or collapse the entire expression to TRUE and FALSE
    // appropriately

    // ==> AND(e1,e2) IFF <simplified expr>
    Theorem rewriteAnd(const Expr& e);
    // ==> OR(e1,...,en) IFF <simplified expr>
    Theorem rewriteOr(const Expr& e);

    // ==> IMPLIES(e1,e2) IFF OR(!e1, e2)
    Theorem rewriteImplies(const Expr& e);
    // ==> IFF(e1,e2) IFF <simplified expr>
    Theorem rewriteIff(const Expr& e);
    // ==> IFF(e1,e2) IFF AND(OR(!e1, e2),OR(e1, !e2))
    Theorem iffToAndOr(const Expr& e);

    // ==> NOT(e) == ITE(e, FALSE, TRUE)
    Theorem NotToIte(const Expr& not_e);
    // ==> Or(e) == ITE(e[1], TRUE, e[0])
    Theorem OrToIte(const Expr& e);
    // ==> And(e) == ITE(e[1], e[0], FALSE)
    Theorem AndToIte(const Expr& e);
    //==> IFF(a,b) == ITE(a, b, !b)
    Theorem IffToIte(const Expr& e);
    // ==> IMPLIES(a,b) == ITE(a, b, TRUE)
    Theorem ImpToIte(const Expr& e);

    // ==> ITE(e, FALSE, TRUE) IFF NOT(e)
    Theorem rewriteIteToNot(const Expr& e);
    // ==> ITE(a, TRUE, b) IFF OR(a, b)
    Theorem rewriteIteToOr(const Expr& e);
    // ==> ITE(a, b, FALSE) IFF AND(a, b)
    Theorem rewriteIteToAnd(const Expr& e);
    // ==> ITE(a, b, !b) IFF IFF(a, b)
    Theorem rewriteIteToIff(const Expr& e);
    // ==> ITE(a, b, TRUE) IFF IMPLIES(a, b)
    Theorem rewriteIteToImp(const Expr& e);

    // ==> Pred(ITE(cond,a,b), x) IFF ITE(cond,Pred(a,x),Pred(b,x))
    Theorem ifLiftRule(const Expr& e, int itePos);

    Theorem varIntroRule(const Expr& phi);
    Theorem andCNFRule(const Theorem& thm);
    Theorem orCNFRule(const Theorem& thm);
    Theorem impCNFRule(const Theorem& thm);
    Theorem iffCNFRule(const Theorem& thm);
    Theorem iteCNFRule(const Theorem& thm);
    Theorem iteToClauses(const Theorem& ite);
    Theorem iffToClauses(const Theorem& iff);


    /////////////////////////////////////////////////////////////////////////
    //// helper functions for CNF (Conjunctive Normal Form) conversion
    /////////////////////////////////////////////////////////////////////////
    private:

    Theorem opCNFRule(const Theorem& thm, int kind,
		      const std::string& ruleName);
    
    Expr convertToCNF(const Expr& v, const Expr & phi);      

    //! Creates a new variable corresponding to the input boolean expression.
    Expr newExprBoundVar(const Expr& e);

    //! checks if phi has v in local cache of opCNFRule, if so reuse v.
    /*! similarly for ( ! ... ! (phi)) */
    Expr findInLocalCache(const Expr& i, 
			  ExprMap<Expr>& localCache,
			  std::vector<Expr>& boundVars);
  }; // end of class CcProofRules

} // end of namespace CVCL

#endif
