/*****************************************************************************/
/*!
 * \file theory.cpp
 * 
 * Author: Clark Barrett
 * 
 * Created: Thu Jan 30 15:11:55 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/


#include <sstream>
#include "theory.h"
#include "debug.h"
#include "theory_core.h"
#include "vcl.h"
#include "common_proof_rules.h"
#include "typecheck_exception.h"
#include "expr_op.h"

using namespace CVCL;
using namespace std;


Theory::Theory(void) : d_theoryCore(NULL) { }


Theory::Theory(VCL* vcl, const string& name)
  : d_theoryCore(vcl->theoryCore()),
    d_commonRules(vcl->theoryCore()->d_commonRules),
    d_name(name), d_theoryUsed(false) {
  d_theoryCore->d_theories.push_back(d_theoryCore);
}


Theory::~Theory(void) { }

Theorem Theory::rewriteThm(const Theorem& thm) {
  DebugAssert(thm.getExpr().isEq() || thm.getExpr().isIff(),
	      "Theory::rewriteThm(): "
	      "Theorem is not an equality or iff: " + thm.toString());
  Theorem thm2(rewrite(thm.getRHS()));
  return d_commonRules->transitivityRule(thm, thm2);
}


Theorem Theory::simplifyOp(const Expr& e) {
  int ar = e.arity();
  if (ar > 0) {
    vector<Theorem> newChildrenThm;
    vector<unsigned> changed;
    for(int k = 0; k < ar; ++k) {
      // Recursively simplify the kids
      Theorem thm = simplify(e[k]);
      if (thm.getLHS() != thm.getRHS()) {
	newChildrenThm.push_back(thm);
	changed.push_back(k);
      }
    }
    if(changed.size() > 0)
      return substitutivityRule(e, changed, newChildrenThm);
  }
  return reflexivityRule(e);
}


void Theory::computeModelTerm(const Expr& e, std::vector<Expr>& v) {
  TRACE("model", "Theory::computeModelTerm(", e, "): is a variable");
  //  v.push_back(e);
}


/*! 
 * \param e is an expression (term or formula).  This function
 * computes the TCC of e which is true iff the expression is defined.
 *
 * This function computes the TCC or predicate of the top-level
 * operator of e, and recurses into children using getTCC(), if
 * necessary.
 *
 * The default implementation is to compute TCCs recursively for all
 * children, and return their conjunction.
 */
Expr Theory::computeTCC(const Expr& e) {
  vector<Expr> kids;
  for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend; ++i)
    kids.push_back(getTCC(*i));
  return (kids.size()>0) ?
    rewriteAnd(andExpr(kids)).getRHS() : trueExpr();
}


ExprManager* Theory::getEM()
{
  return d_theoryCore->d_em;
}


void Theory::registerKinds(Theory* theory, vector<int>& kinds)
{
  vector<int>::const_iterator k;
  vector<int>::const_iterator kEnd;
  for (k = kinds.begin(), kEnd = kinds.end(); k != kEnd; ++k) {
    DebugAssert(d_theoryCore->d_theoryMap.count(*k) == 0,
		"kind already registered: "+getEM()->getKindName(*k)
		+" = "+int2string(*k));
    d_theoryCore->d_theoryMap[*k] = theory;
  }
}


void Theory::registerTheory(Theory* theory, vector<int>& kinds,
			    bool hasSolver)
{
  registerKinds(theory, kinds);
  d_theoryCore->d_theories.push_back(theory);
  if (hasSolver) {
    DebugAssert(!d_theoryCore->d_solver,"Solver already registered");
    d_theoryCore->d_solver = theory;
  }
}


int Theory::getNumTheories()
{
  return d_theoryCore->d_theories.size();
}


Theory* Theory::theoryOf(int kind)
{
  DebugAssert(d_theoryCore->d_theoryMap.count(kind) > 0,
	      "Unknown operator: " + getEM()->getKindName(kind));
  return d_theoryCore->d_theoryMap[kind];
}


Theory* Theory::theoryOf(const Expr& e)
{
  if (e.isNot() || e.isEq()) {
    return theoryOf(e[0]);
  }
  if (!e.isVar()) {
    return theoryOf(e.getKind());
  }
  // Theory of a var is determined by its *base* type, since SUBTYPE
  // may have different base types, but SUBTYPE itself belongs to
  // TheoryCore.
  const Expr& typeExpr = getBaseType(e).getExpr();
  DebugAssert(!typeExpr.isNull(),"Null type");
  DebugAssert(d_theoryCore->d_theoryMap.count(typeExpr.getKind()) > 0,
	      "Unknown operator: " + getEM()->getKindName(typeExpr.getKind()));
  return d_theoryCore->d_theoryMap[typeExpr.getKind()];
}


Theorem Theory::find(const Expr& e)
{
  Theorem thm1 = e.hasFind() ? e.getFind() : reflexivityRule(e);
  const Expr& e1 = thm1.getRHS();
  if (e1 == e || !e1.hasFind()) return thm1;
  Theorem thm2 = find(e1);
  DebugAssert(thm2.getLHS()==e1,"");
  if (thm2.getLHS() != thm2.getRHS()) {
    e.setFind(transitivityRule(thm1,thm2));
    return e.getFind();
  }
  return thm1;
}


const Type& Theory::getType(const Expr& e)
{
  const Type& t = e.lookupType();
  if (!t.getExpr().isNull()) return t;
  DebugAssert(!e.isVar(), "Variables should have a type: "+e.toString());
  Theory* i = theoryOf(e.getKind());
  TRACE("types", "computeType["+i->getName()+"](", e, ") {");
  TRACE("types verbose", "kind = ", getEM()->getKindName(e.getKind()), "");
  i->computeType(e);
  TRACE("types", "computeType["+i->getName()+"] => ", e.lookupType(), " }");
  DebugAssert(!e.lookupType().getExpr().isNull(), "Type not set by computeType");
  return e.lookupType();
}


void Theory::getModelTerm(const Expr& e, std::vector<Expr>& v)
{
  Theory *i = theoryOf(getBaseType(e).getExpr());
  TRACE("model", "computeModelTerm["+i->getName()+"](", e, ") {");
  i->computeModelTerm(e, v);
  TRACE("model", "computeModelTerm[", i->getName(), "] => }");
  DebugAssert(v.size()==0 || v.back() != e, "Primitive var was pushed on "
	      "the stack in computeModelTerm["+i->getName()
	      +"]: "+e.toString());
  
}


Theorem Theory::getModelValue(const Expr& e) {
  return d_theoryCore->getModelValue(e);
}


Expr Theory::getTypePred(const Type& t, const Expr& e)
{
  Expr pred;
  Theory *i = theoryOf(t.getExpr().getKind());
  TRACE("typePred", "computeTypePred["+i->getName()+"]("+t.toString()+", ", e, ") {");
  pred = i->computeTypePred(t, e);
  TRACE("typePred", "computeTypePred["+i->getName()+"] => ", pred, " }");
  return pred;
}

const Expr& Theory::getTCC(const Expr& e)
{
  const Expr& tccCached = e.lookupTCC();
  if (!tccCached.isNull()) return tccCached;
  if(e.isVar()) {
    e.setTCC(trueExpr());
    return trueExpr();
  }
  Theory* i = theoryOf(e.getKind());
  TRACE("tccs", "computeTCC["+i->getName()+"](", e, ") {");
  e.setTCC(i->computeTCC(e));
  TRACE("tccs", "computeTCC["+i->getName()+"] => ", e.lookupTCC(), " }");
  return e.lookupTCC();
}


Type Theory::getBaseType(const Expr& e)
{
  return getBaseType(getType(e));
}

Type Theory::getBaseType(const Type& tp)
{
  const Expr& e = tp.getExpr();
  DebugAssert(!e.isNull(), "Theory::getBaseType(Type(Null))");
  // See if we have it cached
  Type res(e.lookupType());
  if(!res.isNull()) return res;

  DebugAssert(theoryOf(e) != NULL,
	      "Theory::getBaseType(): No theory for the type: "
	      +tp.toString());
  TRACE("types", "getBaseType["+theoryOf(e)->getName()+"](", e, ") {");
  res= theoryOf(e)->computeBaseType(tp);
  TRACE("types", "getBaseType["+theoryOf(e)->getName()+"] => ", res, " }");
  e.setType(res);
  return res;
}


Expr Theory::parseExpr(const Expr& e) {
  TRACE("facts parseExpr", "parseExpr(", e, ") {");
  Expr res(d_theoryCore->parseExpr(e));
  TRACE("facts parseExpr", "parseExpr => ", res, " }");
  return res;
}


bool Theory::leavesAreSimp(const Expr& e)
{
  if (isLeaf(e)) {
    bool res(e.getFind().getRHS() == e);
    IF_DEBUG(if(!res) TRACE("facts leavesAreSimp", "leavesAreSimp(", e, 
			    " = "+e.getFind().getRHS().toString()
			    +") => false"));
    return res;
  }
  for (int k = 0; k < e.arity(); ++k) {
    if (!leavesAreSimp(e[k])) return false;
  }
  return true;
}


Theorem Theory::updateHelper(const Expr& e)
{
  vector<Theorem> newChildrenThm;
  vector<unsigned> changed;
  int k(0), ar(e.arity());
  Theorem thm;
  for (; k < ar; ++k) {
    thm = find(e[k]);
    if (thm.getLHS() != thm.getRHS()) {
      newChildrenThm.push_back(thm);
      changed.push_back(k);
    }
  }
  if (changed.size() == 0) thm = reflexivityRule(e);
  else thm = substitutivityRule(e, changed, newChildrenThm);
  return thm;
}


void Theory::setupTerm(const Expr& e) {
  d_theoryCore->setupTerm(e, this);
}


//! Setup a term for congruence closure (must have sig and rep attributes)
void Theory::setupCC(const Expr& e) {
  TRACE("facts setup", "setupCC["+getName()+"](", e, ") {");
  for (int k = 0; k < e.arity(); ++k) {
    e[k].addToNotify(this, e);
  }
  Theorem thm = reflexivityRule(e);
  e.setSig(thm);
  e.setRep(thm);
  TRACE_MSG("facts setup", "setupCC["+getName()+"]() => }");
}


//! Update a term w.r.t. congruence closure (must be setup with setupCC())
void Theory::updateCC(const Theorem& e, const Expr& d) {
  TRACE("facts update", "updateCC["+getName()+"]("+e.getExpr().toString()+", ",
	d, ") {");
  int k, ar(d.arity());
  const Theorem& dEQdsig = d.getSig();
  if (!dEQdsig.isNull()) {
    const Expr& dsig = dEQdsig.getRHS();
    Theorem thm = updateHelper(d);
    const Expr& sigNew = thm.getRHS();
    if (sigNew == dsig) {
      TRACE_MSG("facts update", "updateCC["+getName()+"]() [no change] => }");
      return;
    }
    dsig.setRep(Theorem());
    const Theorem& repEQsigNew = sigNew.getRep();
    if (!repEQsigNew.isNull()) {
      d.setSig(Theorem());
      enqueueFact(transitivityRule(repEQsigNew, symmetryRule(thm)));
    }
    else {
      for (k = 0; k < ar; ++k) {
	if (sigNew[k] != dsig[k]) {
	  sigNew[k].addToNotify(this, d);
	}
      }
      d.setSig(thm);
      sigNew.setRep(thm);
    }
  }
  TRACE_MSG("facts update", "updateCC["+getName()+"]() => }");
}


//! Rewrite a term w.r.t. congruence closure (must be setup with setupCC())
Theorem Theory::rewriteCC(const Expr& e) {
  const Theorem& rep = e.getRep();
  if (rep.isNull()) return reflexivityRule(e);
  else return symmetryRule(rep);
}


Op Theory::getOpKind(int kind) {
  return getOp(newExpr(getEM(), kind, std::vector<Expr>()));
}

void Theory::setupTerm(const Expr& e, Theory* i) {
  d_theoryCore->setupTerm(e, i);
}


void Theory::addFact(const Theorem& e)
{
  return d_theoryCore->addFact(e);
}


bool Theory::inconsistent()
{
  return d_theoryCore->inconsistent();
}


bool Theory::incomplete()
{
  return d_theoryCore->incomplete();
}


bool Theory::incomplete(vector<string>& reasons)
{
  return d_theoryCore->incomplete(reasons);
}


void Theory::assertEqualities(const Theorem& e)
{
  d_theoryCore->assertEqualities(e);
}


Theorem Theory::simplifyThm(const Expr& e, bool forceRebuild)
{
  TRACE("simplify", ("simplifyThm[" + getName() + "->]("), e, ") {");
  Theorem res(d_theoryCore->simplifyThm(e, forceRebuild));
  TRACE("simplify", "simplifyThm[" + getName() + "] => ", res, "}");
  return res;
}


Theorem Theory::simplify(const Expr& e)
{
  TRACE("simplifyRec", ("simplify[" + getName() + "->]("), e, ") {");
  Theorem res(d_theoryCore->simplify(e));
  TRACE("simplifyRec", "simplify[" + getName() + "] => ", res, "}");
  return res;
}


Theorem Theory::rewriteCore(const Expr& e)
{
  return d_theoryCore->rewriteCore(e);
}


Theorem Theory::rewriteLiteral(const Expr& e) {
  TRACE("rewrite", ("rewriteLiteral[" + getName() + "->]("), e, ") {");
  Theorem res(d_theoryCore->rewriteLiteral(e));
  TRACE("rewrite", "rewriteLiteral[" + getName() + "] => ", res, "}");
  return res;
}


void Theory::enqueueFact(const Theorem& e)
{
  TRACE("facts assertFact", ("enqueueFact[" + getName() + "->]("), e, ")");
  d_theoryCore->enqueueFact(e);
}


void Theory::enqueueEquality(const Theorem& e)
{
  TRACE("facts assertFact", ("enqueueEquality[" + getName() + "->]("), e, ")");
  DebugAssert(e.isRewrite(), "Theory::enqueueEquality("+e.toString()+")");
//   DebugAssert(leavesAreSimp(e.getRHS()),
// 	      "Theory::enqueueEquality("+e.toString()+")");
  d_theoryCore->enqueueEquality(e);
}


void Theory::addSplitter(const Expr& e, int priority) {
  TRACE("facts assertFact", "addSplitter[" + getName() + "->](", e,
	", pri="+int2string(priority)+")");
  d_theoryCore->addSplitter(e, priority);
}


void Theory::setInconsistent(const Theorem& e)
{
  TRACE("facts assertFact", ("setInconsistent[" + getName() + "->]("), e, ")");
  TRACE("conflict", ("setInconsistent[" + getName() + "->]("), e, ")");
  IF_DEBUG(debugger.counter("conflicts from DPs")++);
  d_theoryCore->setInconsistent(e);
}


void Theory::setIncomplete(const string& reason)
{
  TRACE("facts assertFact", "setIncomplete["+getName()+"](", reason, ")");
  d_theoryCore->setIncomplete(reason);
}


Theorem Theory::inconsistentThm()
{
  return d_theoryCore->inconsistentThm();
}


bool Theory::checkSATCore()
{
  return d_theoryCore->checkSATCore();
}


void Theory::setResourceLimit(unsigned limit)
{
  return d_theoryCore->setResourceLimit(limit);
}


bool Theory::isAtomic(const Expr& e)
{
  TRACE("isAtomic", "isAtomic(", e, ") {");
  if (e.validIsAtomicFlag()) {
    TRACE("isAtomic", "isAtomic[cached] => ",
	  e.getIsAtomicFlag()? "true" : "false", " }");
    return e.getIsAtomicFlag();
  }
  if (getType(e).isBool() && !e.isBoolConst()) {
    e.setIsAtomicFlag(false);
    TRACE_MSG("isAtomic", "isAtomic[bool] => false }");
    return false;
  }
  for (int k = 0; k < e.arity(); ++k) {
    if (!isAtomic(e[k])) {
      e.setIsAtomicFlag(false);
      TRACE_MSG("isAtomic", "isAtomic[kid] => false }");
      return false;
    }
  }
  e.setIsAtomicFlag(true);
  TRACE_MSG("isAtomic", "isAtomic[kid] => true }");
  return true;
}


bool Theory::isAtomicFormula(const Expr& e)
{
  TRACE("isAtomic", "isAtomicFormula(", e, ") {");
  if (!getType(e).isBool()) {
    TRACE_MSG("isAtomic", "isAtomicFormula[kid] => false }");
    return false;
  }
  switch(e.getKind()) {
    case FORALL: case EXISTS:
      TRACE_MSG("isAtomic", "isAtomicFormula[quant] => false }");
      return true;
  }
  switch (e.getKind()) {
    case NOT: case AND: case OR: case ITE: case IFF: case IMPLIES:
      TRACE_MSG("isAtomic", "isAtomicFormula[connective] => false }");
      return false;
  }
  for (Expr::iterator k = e.begin(), kend=e.end(); k != kend; ++k) {
    if (!isAtomic(*k)) {
      TRACE_MSG("isAtomic", "isAtomicFormula[kid] => false }");
      return false;
    }
  }
  TRACE_MSG("isAtomic", "isAtomicFormula => true }");
  return true;
}


bool Theory::isLeafIn(const Expr& e1, const Expr& e2)
{
  DebugAssert(isLeaf(e1),"Expected leaf");
  if (e1 == e2) return true;
  if (theoryOf(e2) != this) return false;
  for(Expr::iterator i=e2.begin(), iend=e2.end(); i != iend; ++i)
    if (isLeafIn(e1, *i)) return true;
  return false;
}


Type Theory::boolType()
{
  return Type::typeBool(d_theoryCore->d_em);
}


const Expr& Theory::falseExpr()
{
  return d_theoryCore->d_em->falseExpr();
}


const Expr& Theory::trueExpr()
{
  return d_theoryCore->d_em->trueExpr();
}


Expr Theory::newVar(const string& name, const Type& type)
{
  DebugAssert(!type.isNull(), "Error: defining a variable of NULL type");
  Expr res = resolveID(name);
  Type t;
//   Expr res = lookupVar(name, &t);
  if (!res.isNull()) {
    t = getType(res);
    throw TypecheckException
      ("incompatible redefinition of variable "+name+":\n "
       "already defined with type: "+t.toString()
       +"\n the new type is: "+type.toString());
  }
  res = newVarExpr(getEM(), name);

  // Replace TYPEDEF with its definition
  t = type;
  while(t.getExpr().getKind() == TYPEDEF) {
    DebugAssert(t.arity() == 2, "Bad TYPEDEF: "+t.toString());
    t = t[1];
  }

  // FIXME: why do we need to convert Boolean vars into 0-ary predicates?
  // What's the significance of such a reduction, anyway?
  if(type.isBool()) {
    res.setType(type);
    res = newBoolVarExpr(res);
  }
  // Add the variable for constructing a concrete model
  d_theoryCore->addToVarDB(res);
  // Add the new global declaration
  d_theoryCore->d_globals[name] = res;
  
  DebugAssert(type.getExpr().getKind() != ARROW,"");
  DebugAssert(res.lookupType().isNull(), 
	      "newVarExpr: redefining a variable "
	       + name);
  res.setType(type);
  return res;
}


Expr Theory::newVar(const string& name, const Type& type, const Expr& def)
{
  DebugAssert(!type.isNull(), "Error: defining a variable of NULL type");
  Type t;
  Expr res = resolveID(name);
  if (!res.isNull()) {
    throw TypecheckException
      ("Redefinition of variable "+name+":\n "
       "This variable is already defined.");
  }
  Expr v(newVarExpr(getEM(), name));
  res = newExpr(getEM(), CONSTDEF, v, def);

  // Replace TYPEDEF with its definition
  t = type;
  while(t.getExpr().getKind() == TYPEDEF) {
    DebugAssert(t.arity() == 2, "Bad TYPEDEF: "+t.toString());
    t = t[1];
  }
  v.setType(type);

  // Typecheck
  if(getBaseType(def) != getBaseType(v)) {
    throw TypecheckException("Type mismatch in constant definition:\n"
			     "Constant "+v.toString()+
			     " is declared with type:\n  "
			     +getType(v).toString()
			     +"\nBut the type of definition is\n  "
			     +getType(def).toString());
  }
  DebugAssert(type.getExpr().getKind() != ARROW,"");
  DebugAssert(res.lookupType().isNull(), 
	      "newVarExpr: redefining a variable "
	       + name);

  // Add the new global declaration
  d_theoryCore->d_globals[name] = res;
  
  return res;
}


Op Theory::newFunction(const std::string& name, const Type& type) {
  DebugAssert(!type.isNull(), "Error: defining a variable of NULL type");
  Expr res = resolveID(name);
  Type t;
  if (!res.isNull()) {
    t = getType(res);
    throw TypecheckException
      ("Redefinition of variable "+name+":\n "
       "already defined with type: "+t.toString()
       +"\n the new type is: "+type.toString());
  }
  res = newVarExpr(getEM(), name);
  // Replace TYPEDEF with its definition
  t = type;
  while(t.getExpr().getKind() == TYPEDEF) {
    DebugAssert(t.arity() == 2, "Bad TYPEDEF: "+t.toString());
    t = t[1];
  }
  res.setType(t);
  // Add it to the database of variables for concrete model generation
  d_theoryCore->addToVarDB(res);
  // Add the new global declaration
  d_theoryCore->d_globals[name] = res;
  return getOp(newApplyExpr(getEM(), res, vector<Expr>()));
}


Op Theory::newFunction(const std::string& name, const Type& type,
		       const Expr& def) {
  DebugAssert(!type.isNull(), "Error: defining a variable of NULL type");
  Expr e = resolveID(name);
  Type t;
  if (!e.isNull()) {
    t = getType(e);
    throw TypecheckException
      ("Redefinition of variable "+name+":\n "
       "already defined with type: "+t.toString()
       +"\n the new type is: "+type.toString());
  }
  e = newVarExpr(getEM(), name);
  // Replace TYPEDEF with its definition
  t = type;
  while(t.getExpr().getKind() == TYPEDEF) {
    DebugAssert(t.arity() == 2, "Bad TYPEDEF: "+t.toString());
    t = t[1];
  }
  e.setType(t);
  Expr res(getEM(), CONSTDEF, e, def);
  // Add the new global declaration
  d_theoryCore->d_globals[name] = res;
  return getOp(newApplyExpr(getEM(), res, vector<Expr>()));
}


Expr Theory::newBoundVar(const string& name, const string& uid,
			const Type& type) {
  Expr res = newBoundVarExpr(getEM(), name, uid);
  DebugAssert(type.getExpr().getKind() != ARROW,"");
  DebugAssert(res.lookupType().isNull(), 
	      "newVarExpr: redefining a variable "
	       + name);
  res.setType(type);
  return res;
}


static int boundVarCount = 0;

Expr Theory::addBoundVar(const std::string& name, const Type& type) {
  ostringstream ss;
  ss << boundVarCount++;
  Expr v(newBoundVar(name, ss.str(), type));
  d_theoryCore->d_boundVarStack.push_back(pair<string,Expr>(name,v));
  TRACE("vars", "addBoundVar("+name+", ", type, ") => "+v.toString());
  return v;
}


Expr Theory::addBoundVar(const std::string& name, const Type& type,
			 const Expr& def) {
  Expr res;
  // Without the type, just replace the bound variable with the definition
  if(type.isNull())
    res = def;
  else {
    // When type is given, construct (LETDECL var, def) for the typechecker
    ostringstream ss;
    ss << boundVarCount++;
    res = newExpr(getEM(), LETDECL,
		  newBoundVar(name, ss.str(), type), def);
  }
  d_theoryCore->d_boundVarStack.push_back(pair<string,Expr>(name,res));
  TRACE("vars", "addBoundVar("+name+", "+type.toString()+", ", def,
	") => "+res.toString());
  return res;
}


Expr Theory::lookupVar(const string& name, Type* type)
{
  Expr e = newVarExpr(getEM(), name);
  *type = e.lookupType();
//   if ((*type).isNull()) {
//     e = newApplyExpr(Op(UFUNC, e));
//     *type = e.lookupType();
    if ((*type).isNull()) return Expr();
//   }
  return e;
}


// TODO: reconcile this with parser-driven new type expressions
Type Theory::newTypeExpr(const string& name)
{
  Expr res = resolveID(name);
  if (!res.isNull()) {
    throw TypecheckException
      ("Redefinition of type variable "+name+":\n "
       "This variable is already defined.");
  }
  res = newExpr(getEM(), TYPEDECL, newStringExpr(getEM(), name));
  // Add the new global declaration
  d_theoryCore->d_globals[name] = res;
  return Type(res);
}

//! Create a new type abbreviation with the given name 
Type Theory::newTypeExpr(const std::string& name, const Type& def)
{
  Expr res = resolveID(name);
  if (!res.isNull()) {
    throw TypecheckException
      ("Redefinition of type variable "+name+":\n "
       "This variable is already defined.");
  }
  res = def.getExpr();
  // Add the new global declaration
  d_theoryCore->d_globals[name] = res;
  return Type(res);
}


Expr Theory::resolveID(const std::string& name) {
  TRACE("vars", "resolveID(", name, ") {");
  Expr res; // Result is Null by default
  // First, look up the bound variable stack
  for(vector<pair<string, Expr> >::iterator
	i=d_theoryCore->d_boundVarStack.begin(),
	iend=d_theoryCore->d_boundVarStack.end();
      i!=iend && res.isNull(); ++i) {
    if(i->first == name)
      res = i->second;
  }
  if(res.isNull()) {
    // Next, check in the globals
    map<string,Expr>::iterator i=d_theoryCore->d_globals.find(name),
      iend=d_theoryCore->d_globals.end();
    if(i!=iend)
      res = i->second;
  }
  TRACE("vars", "resolveID => ", res, " }");
  return res;
}


const CDList<Expr>& Theory::getTerms() {
  return d_theoryCore->getTerms();
}


Theorem Theory::subtypePredicate(const Expr& e) {
  return d_theoryCore->subtypePredicate(e);
}


Theorem Theory::skolemize(const Theorem& thm) {
  return d_theoryCore->skolemize(thm);
}


Theorem Theory::varIntroSkolem(const Expr& e) {
  return d_theoryCore->varIntroSkolem(e);
}


void Theory::assignValue(const Expr& t, const Expr& val) {
  TRACE("facts assertFact", "assignValue["+getName()+"](", t, ", "+
	val.toString()+") {");
  d_theoryCore->assignValue(t, val);
  TRACE("facts assertFact", "assignValue[", getName(), "] => }");
}

void Theory::assignValue(const Theorem& thm) {
  TRACE("facts assertFact", "assignValue["+getName()+"](", thm, ") {");
  d_theoryCore->assignValue(thm);
  TRACE("facts assertFact", "assignValue[", getName(), "] => }");
}
