/*****************************************************************************/
/*!
 *\file records_expr.h
 *\brief Special Expr subclasses for records and tuples
 *
 * Author: Sergey Berezin
 *
 * Created: Thu Apr  7 23:07:57 2005
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#ifndef _CVCL__theory_records__records_expr_h_
#define _CVCL__theory_records__records_expr_h_

#include <string>
#include <vector>
#include "expr_value.h"
#include "expr.h"
#include "debug.h"

namespace CVCL {

  //! Record literal: contains list of field names as an attribute
  class ExprRecord: public ExprCC {
    friend class Expr;
    friend class ExprManager;
  private:
    // Special attributes for this type of expression
    std::vector<std::string> d_fields;
    size_t d_mmIdx;

    // Tell ExprManager who we are
    size_t getMMIndex() const { return d_mmIdx; }
    //! Return fields as (std::vector<std::string>*) cast to void*
    const void* getOtherValue(int i) const { return (void*)&d_fields; }
    size_t getOtherValueSize() const { return 1; }
    //! Return fields one at a time
    const std::string& getStringValue(int i) const {
      DebugAssert(i < (int)(d_fields.size()), "ExprRecord::getStringValue("
		  +int2string(i)+", d_fields.size() = "
		  +int2string(d_fields.size())+")");
      return d_fields[i];
    }
    size_t getStringValueSize() const { return d_fields.size(); }
    size_t computeHash() const {
      size_t res = ExprCC::computeHash();
      for(std::vector<std::string>::const_iterator i=d_fields.begin(),
	    iend=d_fields.end(); i!=iend; ++i)
	res = PRIME*res + s_charHash(i->c_str());
      return res;
    }
  public:
    // Constructor
    ExprRecord(size_t mmIdx, ExprManager *em, int kind,
	       const std::vector<std::string>& fields,
	       const std::vector<Expr>& kids, unsigned idx = 0)
      : ExprCC(em, kind, kids, idx), d_fields(fields), d_mmIdx(mmIdx) { }
    // Destructor
    virtual ~ExprRecord() { }

    virtual bool operator==(const ExprValue& ev2) const {
      return (ExprNode::operator==(ev2)
	      && *((const std::vector<std::string>*)getOtherValue(0))
	      == *((const std::vector<std::string>*)ev2.getOtherValue(0)));
    }
    //! Make a clean copy of itself using the given memory manager
    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	ExprRecord(d_mmIdx, em, d_kind, d_fields, d_children, idx);
    }
    ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
			    ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	ExprRecord(d_mmIdx, em, d_kind, d_fields, kids, idx);
    }
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprRecord

  //! Record field access (select, update)
  /*! Contains the field name as an attribute */
  class RecField: public ExprCC {
    friend class Expr;
    friend class ExprManager;
  private:
    // Special attributes for this type of expression
    std::string d_field;
    size_t d_mmIdx;

    // Tell ExprManager who we are
    size_t getMMIndex() const { return d_mmIdx; }

    const std::string& getStringValue(int i) const { return d_field; }
    size_t getStringValueSize() const { return 1; }

    virtual size_t computeHash() const {
      return (s_charHash(d_field.c_str())*PRIME
	      + ExprCC::computeHash());
    }
  public:
    // Constructor
    RecField(size_t mmIdx, ExprManager *em, int kind,
	       const std::string& field,
	       const std::vector<Expr>& kids, unsigned idx = 0)
      : ExprCC(em, kind, kids, idx), d_field(field), d_mmIdx(mmIdx) { }
    // Destructor
    virtual ~RecField() { }

    bool operator==(const ExprValue& ev2) const {
      return (ExprCC::operator==(ev2)
	      && getStringValue(0) == ev2.getStringValue(0));
    }
    //! Make a clean copy of itself using the given memory manager
    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	RecField(d_mmIdx, em, d_kind, d_field, d_children, idx);
    }
    ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
			    ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	RecField(d_mmIdx, em, d_kind, d_field, kids, idx);
    }
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class RecField

  //! Tuple element access (select, update)
  /*! Contains the index value as an attribute */
  class TupleIndex: public ExprCC {
    friend class Expr;
    friend class ExprManager;
  private:
    // Special attributes for this type of expression
    int d_idx;
    size_t d_mmIdx;

    // Tell ExprManager who we are
    size_t getMMIndex() const { return d_mmIdx; }

    size_t computeHash() const {
      return (d_idx*PRIME + ExprCC::computeHash());
    }
    int getIntValue(int i) const { return d_idx; }
    size_t getIntValueSize() const { return 1; }

  public:
    // Constructor
    TupleIndex(size_t mmIdx, ExprManager *em, int kind, int index,
	       const std::vector<Expr>& kids, unsigned idx = 0)
      : ExprCC(em, kind, kids, idx), d_idx(index), d_mmIdx(mmIdx) { }
    // Destructor
    virtual ~TupleIndex() { }

    bool operator==(const ExprValue& ev2) const {
      return (ExprCC::operator==(ev2)
	      && getIntValue(0) == ev2.getIntValue(0));
    }

    //! Make a clean copy of itself using the given memory manager
    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	TupleIndex(d_mmIdx, em, d_kind, d_idx, d_children, idx);
    }
    ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
			    ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	TupleIndex(d_mmIdx, em, d_kind, d_idx, kids, idx);
    }
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class TupleIndex

} // end of namespace CVCL

#endif
